import logging
import random
import string
from flask_sqlalchemy import SQLAlchemy

from portal.database.models import Component, Course, Group, Project, ReviewItem, Role, User

log = logging.getLogger(__name__)


def password_generator(size=16, chars=string.ascii_letters + string.digits):
    """
    Returns a string of random characters, useful in generating temporary
    passwords for automated password resets.

    size: default=8; override to provide smaller/larger passwords
    chars: default=A-Za-z0-9; override to provide more/less diversity

    Credit: Ignacio Vasquez-Abrams
    Source: http://stackoverflow.com/a/2257449
    """
    return ''.join(random.choice(chars) for _ in range(size))


PERM_TEACHER = dict(
    view_course_full=True,
    update_course=True,
    handle_notes_access_token=False,
    assign_roles=True,
    write_roles=True,
    write_groups=True,
    read_groups=True,
    write_projects=True,
    read_projects=True,
    archive_projects=True,
    resubmit_submissions=True,
    evaluate_submissions=True,
    read_submissions_all=True,
    read_all_submission_files=True,
    read_reviews_all=True,
    write_reviews_all=True
    )

PERM_STUDENT = dict(
    view_course_limited=True,
    create_submissions=True,
    read_submissions_own=True,
    read_reviews_own=True,
    write_reviews_own=True
    )

PERM_OWNER = dict(
    view_course_full=True,
    update_course=True,
    handle_notes_access_token=True,
    assign_roles=True,
    write_roles=True,
    write_groups=True,
    read_groups=True,
    write_projects=True,
    read_projects=True,
    archive_projects=True,
    resubmit_submissions=True,
    evaluate_submissions=True,
    read_submissions_all=True,
    read_all_submission_files=True,
    read_reviews_all=True,
    write_reviews_all=True
    )

PERMISSION_TYPES = dict(
    teacher=PERM_TEACHER,
    student=PERM_STUDENT,
    owner=PERM_OWNER,
    )


class DataFactory(object):
    def __init__(self, db: SQLAlchemy):
        self._db = db

    @property
    def session(self):
        return self._db.session

    def create_user(self, username: str, uco=0, name=None, password=None, email=None, admin=False):
        email = email or f"{username}@example.com"
        password = password or "123456"
        user = self.__create_entity(User, uco=uco, username=username, is_admin=admin, email=email)
        user.name = name or f"{username.capitalize()} User"
        user.set_password(password=password)
        return user

    def create_default_admin_user(self, app):
        return self.create_user(
            username=app.config.get('PORTAL_ADMIN_USER_USERNAME', 'admin'),
            password=app.config.get('PORTAL_ADMIN_USER_PASSWORD', password_generator()),
            email=app.config.get('PORTAL_ADMIN_USER_EMAIL', 'admin@example.com'),
            name='Kontr Admin',
            uco=0,
            admin=True,
            )

    def __create_entity(self, klass, *args, **kwargs):
        entity = klass(*args, **kwargs)
        self.session.add(entity)
        log.info(f"[SAMPLE] Create {entity}")
        return entity

    def create_component(self, name: str, secret: str = None, component_type: str = 'executor',
                         ip_addr: str = '127.0.0.1', notes: str = None) -> Component:
        secret = secret or f"{name}_secret"
        notes = notes or f"Component {name} of type {type}"
        return self.__create_entity(Component, name=name, secret=secret,
                                    ip_address=ip_addr, notes=notes, type=component_type)

    def create_course(self, codename, name=None, token=None) -> Course:
        course = self.__create_entity(Course, codename=codename, name=name)
        course.notes_access_token = token or f"{codename}_token"
        return course

    def create_group(self, course: Course, name: str) -> Group:
        return self.__create_entity(Group, course=course, name=name)

    def create_role(self, course: Course, name: str,
                    permissions: dict = None, desc: str = None) -> Role:
        permissions = permissions or {}
        role = self.__create_entity(Role, course=course, name=name)
        role.set_permissions(**permissions)
        role.description = desc or f"{name.capitalize()}'s role"
        return role

    def create_project(self, course, name, test_files_source=None, config=None) -> Project:
        test_files_source = test_files_source or f"git://{name}.repo"
        config = config or {}
        project = self.__create_entity(Project, course=course, name=name,
                                       test_files_source=test_files_source)
        project.set_config(**config)
        return project

    def create_review_item(self, review, author, file="main.c", line=1, content="problem here"):
        return self.__create_entity(ReviewItem, review=review, user=author, file=file,
                                    line=line, content=content)

    def scaffold_course(self, codename: str, name: str = None) -> Course:
        name = name or codename
        course = self.create_course(codename=codename, name=name)
        self.scaffold_role(course, 'teacher')
        self.scaffold_role(course, 'student')
        self.scaffold_role(course, 'owner')
        return course

    def scaffold_role(self, course: Course, role_type: str, name=None) -> Role:
        name = name or role_type
        permissions = PERMISSION_TYPES[role_type]
        role = self.create_role(course=course, name=name, permissions=permissions)
        return role
