import logging
from typing import List

from portal.database.models import User, Course, Group, Submission, Project, Role, Review
from portal.service import general, emails
from portal.service.errors import PortalAPIError, IncorrectCredentialsError
from portal.service.general import find_course, find_user

log = logging.getLogger(__name__)


def __set_user_data(user: User, data: dict, full: bool = False) -> User:
    if (data.get("is_admin") is not None) and full:
        user.is_admin = data['is_admin']
    return general.update_entity(user, data, allowed=['name', 'email', 'uco'])


def create_user(**data) -> User:
    """Creates a new user

    Keyword Args:
        username(str): User's name
        uco(str): User's Uco

    Returns(User): Created user instance
    """
    data['is_admin'] = data.get('is_admin', False)
    new_user = User(username=data['username'])
    __set_user_data(new_user, data, True)
    emails.notify_user(new_user, 'users', 'created', context=dict(username=new_user.username))
    log.info(f"[CREATE] User: {new_user}")
    return new_user


def update_user(user: User, data: dict, full: bool = False) -> User:
    """Updates a user

    Args:
        user(User): User instance
        data(dict): User's data
        full(bool): if true, user can update admin state

    Returns(User): Updated user
    """
    __set_user_data(user, data, full)
    log.info(f"[UPDATE] User {user.id} to {user}.")
    return user


def update_user_password(client: User, data: dict, user: User) -> User:
    """Updates a user's password. If requested by an admin, old password is not required.

    Args:
        client(User): Who has requested the users password's change
        data(dict): Passwords context - new password and old password, if necessary
        user(User): User whose password should be updated

    Returns(User): User instance
    """
    log.info(f"[UPDATE] User password: {user.id} by {client.id}")
    if not data.get('new_password'):
        raise PortalAPIError(400, 'missing new password at password change')
    if client == user and not user.is_admin:
        if not data.get('old_password'):
            raise PortalAPIError(400, 'missing old password for password update')
        old_password = data['old_password']
        if not user.verify_password(old_password):
            raise IncorrectCredentialsError()
    user.set_password(data['new_password'])
    general.write_entity(user)
    emails.notify_user(user, 'users', 'password_update', context=dict(username=user.username))
    return user


def delete_user(user: User):
    """Deletes a user

    Args:
        user(User): User instance
    """
    log.info(f"[DELETE] User({user.username}): {user.id} .")
    general.delete_entity(user)


def find_users(ids: List[str]) -> List[User]:
    """Finds all users based on their ids

    Args:
        ids(List[str]): List of user ids

    Returns(List[User]): List of users
    """
    return [find_user(i) for i in ids]


def find_users_filtered(course_id: str = None, group_id: str = None) -> List[User]:
    """Find all users, optionally filtered by course and group
    Args:
        course_id(str): Course id (optional)
        group_id(str): Group id (optional)

    Returns(List[User]): Filtered list of  users
    """
    users = User.query
    if course_id:
        users = users.join(User.roles) \
            .join(Course).filter((Course.id == course_id) | (Course.codename == course_id))
    if group_id:
        users = users.join(User.groups).filter((Group.id == group_id) | (Group.name == group_id))

    return users.all()


def find_groups_filtered(user: User, course_id: str = None) -> List[Group]:
    """Find all groups for the user, optionally filtered by course.

    Args:
        user(User): User instance
        course_id(str): Course id (optional)

    Returns(List[Group]): List of all groups
    """
    if course_id:
        return user.get_groups_in_course(find_course(course_id))
    return user.groups


def find_roles_filtered(user: User, course_id: str = None) -> List[Role]:
    """Get all roles for the user, optionally filtered by course

    Args:
        user(User): User instance
        course_id(str): Course id (optional)

    Returns(List[Role]):
    """
    roles = user.roles
    if course_id:
        roles = user.get_roles_in_course(find_course(course_id))
    return roles


def find_reviews(user: User) -> List[Review]:
    """Gets reviews the user has contributed to.

    Args:
        user(User): User instance

    Returns(List[Review]):
    """
    reviews = []
    for item in user.review_items:
        if item not in reviews:
            reviews.append(item.review)
    return list(reviews)


def find_submissions_filtered(user: User, course_id: str = None,
                              project_ids: List[str] = None) -> List[Submission]:
    """Get all submissions of a user, optionally filtered by course and project.

    Args:
        user(User):
        course_id(str): Course id (optional)
        project_ids(List[str]): List of project ids (optional)

    Returns(List[Submission]): submissions in the course and projects
    """
    submissions = Submission.query.filter_by(user=user)
    # user.submissions cannot be used as the base query
    if course_id:
        submissions = submissions.join(Project)
        if project_ids:
            submissions = submissions.filter(
                (Project.id.in_(project_ids)) | (Project.name.in_(project_ids)))
        submissions = submissions.join(Course).filter(
            (Course.id == course_id) | (Course.codename == course_id))
    return submissions
